// $Id$
/** *****************************************************************************************
@doxygen

@file tsonewi.cpp

@brief Abfragen von Dallas DS1820 One-Wire-Thermometer.

Siehe auch <a href=../../../deutsch/lq/tsonewi.htm>../../../deutsch/lq/tsonewi.htm</a>
<br>
<a href=../../../../../etronik/sensor/DS18S20.pdf>../../../../../etronik/sensor/DS18S20.pdf</a>

@author HJHerbert

@version 2014-08-16 - noch ohne CRC-Test

********************************************************************************************* */
#include <heport.h>
#include <helq.h>	// outportb()
#include <hegettim.h>	// delaytick()
#include <tshar.h>
#include <tsonewi.h>

/*
<br>ROM Kommandos sind:
<br>F0 - Suche ROM : Muss fuer jeden Slave wiederholt werden.
<br>33 - Lese ROM : Den ROM-Inhalt des einzigen Slave auslesen: 8 Bytes einlesen
<br>55 - ROM Vergleichen : 8 Bytes folgen
<br>CC - den einzigen Slave ansprechen
<br>EC - Grenzwertueberschreitungen suchen.
*/

#if defined( __TURBOC__ ) || defined ( _MSC_VER )
// Unter DOS / Win95 das printerport verwenden
//#define LPT		// LPT-Port verwenden
#endif

#ifdef AVR

// Im makefile stand EINE der Definitionen:
//                 PORTA                     PORTB                     PORTC                     PORTD
// Bit 0 -DTSONEWI_BITA=0x01, oder -DTSONEWI_BITB=0x01, oder -DTSONEWI_BITC=0x01, oder -DTSONEWI_BITD=0x01
// Bit 1 -DTSONEWI_BITA=0x02, oder -DTSONEWI_BITB=0x02, oder -DTSONEWI_BITC=0x02, oder -DTSONEWI_BITD=0x02
// Bit 2 -DTSONEWI_BITA=0x04, oder -DTSONEWI_BITB=0x04, oder -DTSONEWI_BITC=0x04, oder -DTSONEWI_BITD=0x04
// Bit 3 -DTSONEWI_BITA=0x08, oder -DTSONEWI_BITB=0x08, oder -DTSONEWI_BITC=0x08, oder -DTSONEWI_BITD=0x08
// ...
// Bit 7 -DTSONEWI_BITA=0x80, oder -DTSONEWI_BITB=0x80, oder -DTSONEWI_BITC=0x80, oder -DTSONEWI_BITD=0x80
#if defined(TSONEWI_BITA)
#define TSONEWI_PIN PINA
#define TSONEWI_PORT PORTA
#define TSONEWI_DDR DDRA
#define TSONEWI_BIT TSONEWI_BITA
#elif defined(TSONEWI_BITB)
#define TSONEWI_PIN PINB
#define TSONEWI_PORT PORTB
#define TSONEWI_DDR DDRB
#define TSONEWI_BIT TSONEWI_BITB
#elif defined(TSONEWI_BITC)
#define TSONEWI_PIN PINC
#define TSONEWI_PORT PORTC
#define TSONEWI_DDR DDRC
#define TSONEWI_BIT TSONEWI_BITC
#elif defined(TSONEWI_BITD)
#define TSONEWI_PIN PIND
#define TSONEWI_PORT PORTD
#define TSONEWI_DDR DDRD
#define TSONEWI_BIT TSONEWI_BITD
#elif defined(TSONEWI_BITE)
#define TSONEWI_PPIN PINE
#define TSONEWI_PORT PORTE
#define TSONEWI_DDR DDRE
#define TSONEWI_BIT TSONEWI_BITE
#endif


// #define TSONEWI_BIT 0x40		// 0x40 = PD6  
#define SETZE0 TSONEWI_PORT &= ~TSONEWI_BIT ;	/* 0V ausgeben */ \
		TSONEWI_DDR |= TSONEWI_BIT ;			/* Aktiv */

#define SETZE1 TSONEWI_PORT |= TSONEWI_BIT ;		/* 5V ausgeben */ \
		TSONEWI_DDR |= TSONEWI_BIT ;				/* Aktiv */

#define PASSIV TSONEWI_DDR &= ~TSONEWI_BIT ;		/* Nicht Aktiv */ \
		TSONEWI_PORT &= ~TSONEWI_BIT ;		/* nicht mit Pull-Up, der ist extern */

#define EINGANG ((TSONEWI_PIN & TSONEWI_BIT)!=0)
#else
// DOS; WIN95
#ifdef LPT        // PCPARTWI
u8 lpt ;
#define OMASK 0x01
#define SETZE0 lpt |=  OMASK ;	/* 0V ausgeben: Transistor ansteuern */ \
		outportb( 0x378, lpt );
#define SETZE1 lpt &= ~OMASK ;	/* 5V ausgeben: Transistor nicht ansteuern */ \
		outportb( 0x378, lpt );
#define PASSIV lpt &= ~OMASK ;	/* 5V ausgeben, der Ausgang ist so aehnlich wie ein offener Kollektor */ \
		outportb( 0x378, lpt );
#define EINGANG ( inportb( 0x379 ) & 0x10 )
#else
// Linux, Windows
#define SETZE0
#define SETZE1
#define PASSIV
#define EINGANG 0
#endif	// LPT
#endif // AVR

// Alle globalen statischen Variablen hier einmal anlegen
u8 TSOneWire::bu[BULEN];			// 9 empfangene Bytes
CRC8 TSOneWire::crc8;					// berechnete Prüfsumme CRC8
u8 TSOneWire::bus[SERNUMLEN];		// 8 empfangene Bytes
PGM_P TSOneWire::fehler ;			// Fehler - String. Zeigt auf Programmspeicher
u8 TSOneWire::sernum[TSOneWire_N][SERNUMLEN];	// Seriennummern der N Temperatursensoren


/*!
 * Loese einen elektronischen Reset aus, initialisiere den DS1820 und erkenne die Rueckmeldung.
 * <br>Ergebnis: 
 * <ul><li>( TSOneWire::fehler = 0 ) - es ist kein Fehler aufgetreten.</li>
 * <li>( TSOneWire::fehler = text ) - es ist ein Fehler aufgetreten. Ansteuerung muss abgebrochen werden.</li>
 * </ul>
 */
void TSOneWire::Reset(void)
{
u32 starttick ;
u8 i ;


SETZE1 ;
for ( i = 0 ; i < 8 ; ++i )
	{
	bu[i] = 0x55 ;
	}
tsdelay( sectotick( 0.002000 ) );	// 

SETZE0 ;
starttick = gettick();
for(;;)
	{
	if ( EINGANG )
 		{				// Nicht Low
		fehler = PSTR("Reset(): Ich kann die Leitung nicht nach Low ziehen") ;
		return ;
		}
	if ( difftick( starttick, gettick() ) >= sectotick( 0.000480 ) ) 
		{				// lange genug fuer RESET
		break ; 
		}
	}
PASSIV ;

starttick = gettick();
for(;;)
	{
	if ( !EINGANG ) break ;	// Low
	if ( difftick( starttick, gettick() ) >= sectotick( 0.000480 ) )
		{
		fehler = PSTR("Reset(): Leitung wird nicht Low") ;
		return ;
		}
	}
// 18DS20 hat 0 geschickt
for(;;)
	{
	if ( EINGANG ) break ;
	if ( difftick( starttick, gettick() ) >= sectotick( 0.000880 ) )
		{
		fehler = PSTR("Reset(): Leitung am Ende bleibt Low") ;
		return ;
		}
	}
// 18DS20 hat wieder 1 geschickt
//	reset_dt1=59 Ticks 0.000029 s
//  reset_dt2=289 Ticks 0.000144 s
// INIT-Impuls wurde beantwortet
delaytick( sectotick( 0.000008 ) );
fehler = 0 ;
// 0.000008 : CC 64 bringt FE, sonst nichts
// 0.000520 : CC 64 bringt FE, sonst nichts
return ;
}


/*!
 * Schreibe ein 0-Bit
 */
void TSOneWire::Schreibe0(void) // > 60 + 5 us
{
SETZE0 ;
tsdelay( sectotick( 0.000115 ) );		// 60 ... 120
SETZE1 ;
delaytick( sectotick( 0.000015 ) );	
}


/*!
 * Schreibe ein 1-Bit
 */
void TSOneWire::Schreibe1(void) // 10 + 55 us
{
SETZE0 ;
delaytick( sectotick( 0.000015 ) );	// 1 .. 15 us
SETZE1 ;
tsdelay( sectotick( 0.000115 ) );	// 60-t0 ... unbegrenzt
}



/*!
 * Lies ein Bit vom DS1820 ein.
 *
 * <br>Ergebnis: 
 * <ul><li>( TSOneWire::fehler = 0 ) - es ist kein Fehler aufgetreten.</li>
 * <li>( TSOneWire::fehler = text ) - es ist ein Fehler aufgetreten. Ansteuerung muss abgebrochen werden.</li>
 * </ul>
 * @return - das eingelesene Bit, 0 oder 1. Im Fehlerfall ist da 2 oder 3.
 */
FUNCT u8 TSOneWire::Lese(void)
{
u32 starttick ;

delaytick( sectotick( 0.000035 ) );	// 1...unendlich
SETZE0 ;		// Sendeaufforderung
delaytick( sectotick( 0.000001 ) );	// 1  us

PASSIV ;
delaytick( sectotick( 0.000001 ) );	// 15...60 us

starttick = gettick();

if ( EINGANG )
	{	// Eine 1 steht an
	for(;;)
		{
		if ( !EINGANG ) // Fehler
			{
			fehler = PSTR("Lese(): anstehende 1 nicht stabil");
			return 3 ;
			}
		if ( difftick( starttick, gettick() ) >= sectotick( 0.000080 ) )
			{	// Das ist kein Fehler !!
			return 1 ;	// Die 1 stand stabil an
			}
		}
	}
else
	{
	for(;;)
		{
		if ( EINGANG ) // Fehler
			{	// Gesehen bei CC 64
			fehler = PSTR("Lese(): anstehende 0 nicht stabil");
			return 2 ;
			}
		if ( difftick( starttick, gettick() ) >= sectotick( 0.000005 ) )
			{
			break ;	// Die 0 stand stabil an
			}
		}
// Und jetzt warten, bis die 0 weggeht
	for(;;)
		{
		if ( EINGANG ) // Die ankommende 0 ist weggegangen
			{
			return 0 ;
			}
		if ( difftick( starttick, gettick() ) >= sectotick( 0.000090 ) )
			{			// Wartezeit ist angelaufen
			fehler = PSTR("Lese(): anstehende 0 geht nicht weg");
			return 4 ;	// Die 0 geht nicht mehr weg
			}
		}
		
	}
}



/*!
 * Lies ein Byte vom DS1820 ein.
 *
 * <br>Ergebnis: 
 * <ul><li>( TSOneWire::fehler = 0 ) - es ist kein Fehler aufgetreten.</li>
 * <li>( TSOneWire::fehler = text ) - es ist ein Fehler aufgetreten. Ansteuerung muss abgebrochen werden.</li>
 * </ul>
 * @return - das eingelesene Byte.
 */
FUNCT u8 TSOneWire::Hole8Bits(void)
{
u8 b ;
u8 i ;
b = 0 ;
for ( i = 8+1 ; --i ; )
	{	// i = 8,7,6,5,4,3,2,1
	switch( Lese() )
		{
	case 0 : b >>= 1 ; break ;
	case 1 : b >>= 1 ; b |= 0x80 ; break ;
	default :  ; // hat fehler gesetzt
		}
	tsnext("hole");
	}
return b ;
}




/*!
 * Sende ein Byte an den DS1820 ein.
 *
 * <br>Ergebnis: 
 * <ul><li>( TSOneWire::fehler = 0 ) - es ist kein Fehler aufgetreten.</li>
 * <li>( TSOneWire::fehler = text ) - es ist ein Fehler aufgetreten. Ansteuerung muss abgebrochen werden.</li>
 * </ul>
 * @param b - Das zu sendende Byte.
 */
FUNCT void TSOneWire::Sende8Bits(u8 b)
{
u8 i ;

tsdelay( sectotick( 0.000160 ) );	// 

for ( i = 8+1 ; --i ; )
	{	// i = 8,7,6,5,4,3,2,1
	if ( b & 0x01 ) 
		{
		Schreibe1() ;
		}
	else
		{
		Schreibe0() ;
		}
	b >>= 1 ;
	tsnext("sende");
	}
// i = 0
return ;
}



/*!
 * Sende 8 Bytes an den DS1820 ein.
 *
 * <br>Ergebnis: 
 * <ul><li>( TSOneWire::fehler = 0 ) - es ist kein Fehler aufgetreten.</li>
 * <li>( TSOneWire::fehler = text ) - es ist ein Fehler aufgetreten. Ansteuerung muss abgebrochen werden.</li>
 * </ul>
 * @param p - Zeiogt auf 8 Bytes.
 */
FUNCT void TSOneWire::Sende8Bytes(u8 *p)
{
u8 i ;

for ( i = 0 ; i<8 ; ++i)
	{
	Sende8Bits( p[i] );
	if ( fehler ) break ;
	}
// i = 0
return ;
}


/*!
 * Hole 8 Bytes vom DS1820 herein.
 *
 * <br>Ergebnis: 
 * <ul><li>( TSOneWire::fehler = 0 ) - es ist kein Fehler aufgetreten.</li>
 * <li>( TSOneWire::fehler = text ) - es ist ein Fehler aufgetreten. Ansteuerung muss abgebrochen werden.</li>
 * </ul>
 * Das Ergebnis ist in TSOneWire::bu[0] ... TSOneWire::bu[7]
 */
FUNCT void TSOneWire::HoleNBytes(u8 *p,u8 n)
{
u8 i ;

for ( i = 0 ; i < n ; ++i )
	{
	bu[i] = Hole8Bits();	// Lese noch eine Zahl
	if ( fehler ) break ;
	}

for ( i = 1 ; i < n ; ++i )
	{
	if ( bu[i] != 0xff ) break ;	// Gute Werte
	}
	if ( i == n )
	{	// Alle Werte = 0xff
	fehler = PSTR("Hole8Bytes(): nur ff gekommen");
	return ;
	}
// Pruefe CRC8
crc8.Start(0);
for ( i = 0 ; i < n-1 ; ++i )
	{
	crc8.NochEin(p[i]);
	}
if ( p[n-1] != crc8.Ergebnis() )
	{
	fehler = PSTR("Hole8Bytes(): CRC8-Fehler");
	return ;
	}
// Kein Fehler
return ;

}


/*!
 * Waehle einen DS1820-Sensor aus, oder verwende den einzig angeschlossenen ohne Auswahl.
 *
 * <br>Ergebnis: 
 * <ul><li>( TSOneWire::fehler = 0 ) - es ist kein Fehler aufgetreten.</li>
 * <li>( TSOneWire::fehler = text ) - es ist ein Fehler aufgetreten. Ansteuerung muss abgebrochen werden.</li>
 * </ul>
 * @param Kanal - <ul><li>255 : den einzig angeschlossenen auswählen.</li>
 * <li>0..3 : Die Seriennummer des Sensors hinausschicken, die in den 8 Bytes TSOneWire::sernum[Kanal][0] ... TSOneWire::sernum[Kanal][7] hinterlegt ist.</li>
 * </ul>
 */
FUNCT void TSOneWire::SelectROM( u8 Kanal )
{
Reset();
if ( fehler ) { return; } 
if ( Kanal == 0xff )
	{		// Den einzigen, der angeschlossen ist
	Sende8Bits( 0xCC );		// Skip ROM
	// if ( fehler ) { return ; }
	}
else if ( Kanal < TSOneWire_N )
	{
	Sende8Bits( 0x55 );		// Select ROM
	if ( fehler ) { return ; } 
	Sende8Bytes( sernum[Kanal] );		// Die Seriennummer senden
	// if ( fehler ) { return ; }
	}
}


/*!
 * Hole eine Temperatur von einen DS1820-Sensor.
 *
 * <br>Ergebnis: 
 * <ul><li>( TSOneWire::fehler = 0 ) - es ist kein Fehler aufgetreten. Der Wert in 1/2 grad C wird zurueckgegeben.
 * <br>Beispiel:
 * <br>return  0 = 0.0 grad C
 * <br>return  1 = 0.5 grad C
 * <br>return 61 = 30.5 grad C
 * <br>return -2 = -1.0 grad C
 * </li>
 * <li>( TSOneWire::fehler = text ) - es ist ein Fehler aufgetreten. Ansteuerung muss abgebrochen werden.</li>
 * </ul>
 * @param Kanal - <ul><li><b>255</b> : den einzig angeschlossenen auswählen.</li>
 * <li>0..3 : Die Seriennummer des Sensors, die in TSOneWire::sernum[Kanal][0] ... TSOneWire::sernum[Kanal][7] hinterlegt ist.</li>
 * </ul>
 */
FUNCT s16 TSOneWire::Temperatur(u8 Kanal)
{
s16 ganzegrade;
SelectROM( Kanal );
if ( fehler ) { return -32768; } 
Sende8Bits( 0x44 );	// Wandlung anstarten [Convert T]

tsdelay( sectotick( 0.800 ) );	// Wandlung abwarten

SelectROM( Kanal );
if ( fehler ) { return -32768; } 
Sende8Bits( 0xBE );	// Lese alle 9 Zellen
if ( fehler ) { return -32768; } 
HoleNBytes( bu, 9 );		// Lese alle 9 Bytes
// Jetzt sollte der Temperaturwert sein in bu[0] und bu[1]
ganzegrade = (*(s16 *)(&bu[0])) >>1 ;	// Das ist etwas anderes als /2 !!!
return div_sl_sw( mul_sw_sw( (ganzegrade+1)*bu[7]-bu[6], 100 ) , bu[7] ) -25 ;

}
